package com.perforce.bumble;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import difflib.Delta;
import difflib.Delta.TYPE;
import difflib.DiffUtils;
import difflib.Patch;

public class Diff {

	private static Logger logger = LoggerFactory.getLogger(Diff.class);

	private final File original;
	private final File revised;

	public Diff(File original, File revised) {
		this.original = original;
		this.revised = revised;
	}

	public List<DiffModel> getDeltas() throws IOException {

		final String origName = original.getName();
		final String revdName = revised.getName();
		final List<String> origFileLines = fileToLines(original);
		final List<String> revdFileLines = fileToLines(revised);

		List<DiffModel> result = new ArrayList<>();

		for (String l : origFileLines) {
			DiffModel model = new DiffModel(null, l);
			result.add(model);
		}

		final Patch<String> patch = DiffUtils.diff(origFileLines, revdFileLines);

		int offset = 0;
		for (Delta<String> delta : patch.getDeltas()) {
			TYPE type = delta.getType();

			logger.info(type + ":" + delta.getOriginal().getPosition() + "," + delta.getRevised().getPosition());

			if (type == TYPE.DELETE || type == TYPE.CHANGE) {
				int i = delta.getOriginal().getPosition() + offset;

				if (type == TYPE.DELETE) {
					marker(delta, result, i);
					offset++;
					i++;
				}

				List<String> delete = delta.getOriginal().getLines();
				for (String line : delete) {
					DiffModel model = result.get(i);
					if (model.getLine().equals(line)) {
						model.setType("DELETE");
						i++;
					} else {
						logger.error("Error line different:" + line);
					}
				}
			}

			if (type == TYPE.INSERT || type == TYPE.CHANGE) {
				int i = delta.getOriginal().getPosition() + offset;

				marker(delta, result, i);
				offset++;
				i++;

				List<String> insert = delta.getRevised().getLines();
				for (String line : insert) {
					DiffModel model = new DiffModel("INSERT", line);
					result.add(i, model);
					i++;
					offset++;
				}
			}
		}

		int o = 0;
		int r = 0;
		for (DiffModel line : result) {
			switch (line.getType()) {
			case "insert":
				r++;
				break;
			case "delete":
				o++;
				break;
			case "marker":
				break;
			default:
				o++;
				r++;
				break;
			}
			line.setOrig(o);
			line.setRevd(r);
		}

		return result;
	}

	private void marker(Delta<String> delta, List<DiffModel> result, int offset) {
		int c = 3;
		c = (offset > c) ? c : offset;

		int orig = delta.getOriginal().getPosition() + 1 - c;
		int osize = delta.getOriginal().size() + (2 * c);
		int revd = delta.getRevised().getPosition() + 1 - c;
		int rsize = delta.getRevised().size() + (2 * c);

		DiffModel marker = new DiffModel("MARKER", "@@ -" + orig + "," + osize + " +" + revd + "," + rsize + " @@");
		result.add(offset - c, marker);
	}

	private List<String> fileToLines(File file) throws IOException {
		final List<String> lines = new ArrayList<String>();
		String line;
		final BufferedReader in = new BufferedReader(new FileReader(file));
		while ((line = in.readLine()) != null) {
			lines.add(line);
		}
		in.close();
		return lines;
	}

	public String getUnified() throws IOException {
		final String origName = original.getName();
		final String revdName = revised.getName();
		final List<String> origFileLines = fileToLines(original);
		final List<String> revdFileLines = fileToLines(revised);

		List<DiffModel> result = new ArrayList<>();

		for (String l : origFileLines) {
			DiffModel model = new DiffModel(null, l);
			result.add(model);
		}

		StringBuffer sb = new StringBuffer();
		final Patch<String> patch = DiffUtils.diff(origFileLines, revdFileLines);
		List<String> diff = DiffUtils.generateUnifiedDiff(origName, revdName, origFileLines, patch, 3);
		for (String line : diff) {
			sb.append(line + "\n");
		}

		return sb.toString();
	}
}
